"""EditMenu provides the edit menu for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow

from attribox import AttriBox
from ezside.windows.menus import AbstractMenu, Action


class EditMenu(AbstractMenu):
  """EditMenu provides the edit menu for the main application window."""

  undo = AttriBox[Action]('undo')
  redo = AttriBox[Action]('redo')
  cut = AttriBox[Action]('cut')
  copy = AttriBox[Action]('copy')
  paste = AttriBox[Action]('paste')
  delete = AttriBox[Action]('delete')
  select_all = AttriBox[Action]('select_all')

  def __init__(self, parent: QMainWindow) -> None:
    """EditMenu provides the edit menu for the main application window."""
    AbstractMenu.__init__(self, parent)
    self.setTitle('Edit')

  def appendActions(self) -> None:
    """Appends the actions without setting them up"""
    self.addAction(self.undo)
    self.addAction(self.redo)
    self.addSeparator()
    self.addAction(self.cut)
    self.addAction(self.copy)
    self.addAction(self.paste)
    self.addAction(self.delete)
    self.addSeparator()
    self.addAction(self.select_all)

  def setupActions(self, ) -> None:
    """Sets up the actions for the edit menu."""
    self.undo.setupAction()
    self.redo.setupAction()
    self.cut.setupAction()
    self.copy.setupAction()
    self.paste.setupAction()
    self.delete.setupAction()
    self.select_all.setupAction()



"""ControlWidget provides a collection of buttons for controlling
start/stop and timed behaviour."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QPushButton, QVBoxLayout, QHBoxLayout, \
  QBoxLayout
from attribox import AttriBox
from icecream import ic
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from ezside.core import Tight
from ezside.widgets import BaseWidget, VSpacer, HSpacer, AbstractSpacer


class PushButton(QPushButton):
  """PushButton provides a button with a signal."""

  def __init__(self, *args, **kwargs) -> None:
    QPushButton.__init__(self, *args, **kwargs)
    self.setSizePolicy(Tight, Tight)
    self.__inner_text__ = None
    textKeys = stringList("""text, msg, message, label, title""")
    for key in textKeys:
      if key in kwargs:
        val = kwargs[key]
        if isinstance(val, str):
          self.__inner_text__ = val
          break
        e = typeMsg('text', val, str)
        raise TypeError(e)
    else:
      for arg in args:
        if isinstance(arg, str):
          self.__inner_text__ = arg
          break
      else:
        self.__inner_text__ = 'Click Me!'

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    pass


class AbstractControl(BaseWidget):
  """ControlWidget provides a collection of buttons for controlling
  start/stop and timed behaviour."""

  start = Signal()
  stop = Signal()
  pause = Signal()

  startButton = AttriBox[PushButton]('Start')
  pauseButton = AttriBox[PushButton]('Pause')
  stopButton = AttriBox[PushButton]('Stop')

  horizontalSpacer = AttriBox[HSpacer]()
  verticalSpacer = AttriBox[VSpacer]()
  horizontalLayer = AttriBox[QHBoxLayout]()
  verticalLayout = AttriBox[QVBoxLayout]()

  __layout_orientation__ = None

  @abstractmethod
  def getSpacer(self) -> AbstractSpacer:
    """The getSpacer method returns a spacer widget."""

  @abstractmethod
  def getLayout(self) -> QBoxLayout:
    """The getLayout method returns the layout type."""

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.startButton.initUi()
    self.getLayout().addWidget(self.startButton)
    self.pauseButton.initUi()
    self.getLayout().addWidget(self.pauseButton)
    self.stopButton.initUi()
    self.getLayout().addWidget(self.stopButton)
    self.setLayout(self.getLayout())
    self.getLayout().addWidget(self.getSpacer())
    self.connectActions()

  def connectActions(self) -> None:
    """The connectActions method connects the actions of the window."""
    self.startButton.clicked.connect(self.start.emit)
    self.pauseButton.clicked.connect(self.pause.emit)
    self.stopButton.clicked.connect(self.stop.emit)


class VerticalAbstractControl(AbstractControl):
  """VerticalControl provides a collection of buttons for controlling
  start/stop and timed behaviour."""

  def getSpacer(self) -> AbstractSpacer:
    """The getSpacer method returns a spacer widget."""
    return self.verticalSpacer

  def getLayout(self) -> QBoxLayout:
    """The getLayout method returns the layout type."""
    return self.verticalLayout


class HorizontalAbstractControl(AbstractControl):
  """HorizontalControl provides a collection of buttons for controlling
  start/stop and timed behaviour."""

  def getSpacer(self) -> AbstractSpacer:
    """The getSpacer method returns a spacer widget."""
    return self.horizontalSpacer

  def getLayout(self) -> QBoxLayout:
    """The getLayout method returns the layout type."""
    return self.horizontalLayer



"""DataWidget provides a place to put the DataView."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Signal, Slot
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout
from attribox import AttriBox

from ezside.widgets import BaseWidget, WhiteNoise, DataView
from settings import Default


class DataWidget(BaseWidget):
  """DataWidget provides a place to put the DataView."""

  __fallback_num_points__ = Default.numPoints

  start = Signal()
  stop = Signal()
  pause = Signal()

  baseLayout = AttriBox[QVBoxLayout]()
  dataView = AttriBox[DataView]()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(64, 64)
    self.dataView.initUi()
    self.baseLayout.addWidget(self.dataView)
    self.setLayout(self.baseLayout)

  @Slot()
  def refresh(self) -> None:
    """The refresh method refreshes the data."""
    self.dataView.refresh()



"""TestWindow tests widgets one at a time"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QVBoxLayout, QMainWindow
from attribox import AttriBox

from ezside.windows import BaseWindow
from ezside.widgets import BaseWidget, DataWidget, Equalizer


# Example usage


class TestWindow(BaseWindow):
  """TestWindow tests widgets one at a time"""

  baseWidget = AttriBox[BaseWidget]()
  baseLayout = AttriBox[QVBoxLayout]()
  wrapView = AttriBox[DataWidget]()
  testWidget = AttriBox[Equalizer]()

  def __init__(self, *args, **kwargs) -> None:
    BaseWindow.__init__(self, *args, **kwargs)

  def connectActions(self, ) -> None:
    """Connects the actions to the slots."""
    BaseWindow.connectActions(self)

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(256, 256)
    # self.wrapView.initUi()
    self.testWidget.initUi()
    self.baseLayout.addWidget(self.testWidget)
    self.baseWidget.setLayout(self.baseLayout)
    self.setCentralWidget(self.baseWidget)
    BaseWindow.initUi(self, )

  def show(self, ) -> None:
    """Shows the window."""
    self.initUi()
    self.connectActions()
    QMainWindow.show(self)

  def debug1Func(self, ) -> None:
    """Debug function."""
    print('Repaints the window')
    self.testWidget.update()

  def debug2Func(self, ) -> None:
    """Debug function."""
    print('Repaints the window')
    self.update()
    print(self.testWidget.geometry())



"""Attempts to load the icon"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from PIL import Image
from vistutils.text import monoSpace
from vistutils.waitaminute import typeMsg

from ezside.windows.menus.icons import getIconPath


def loadIcon(name: str) -> Image:
  """Attempts to load the icon"""
  iconPath = getIconPath(name)
  if iconPath is not None:
    if isinstance(iconPath, str):
      if os.path.exists(iconPath):
        if os.path.isfile(iconPath):
          return Image.open(iconPath)
        e = """Expected a file at '%s', but found a directory!"""
        raise IsADirectoryError(monoSpace(e % iconPath))
      e = """No file at path: '%s'!"""
      raise FileNotFoundError(monoSpace(e % iconPath))
    e = typeMsg('iconPath', iconPath, str)
    raise TypeError(e)
  e = """Given name: '%s' not recognized as the name of an icon!"""
  raise NameError(monoSpace(e % name))



"""EnumMeta class for the EZNum base class"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable, Any

from icecream import ic
from vistutils.text import monoSpace
from eznum import auto


class EZnumMeta(type):
  """EnumMeta provides the metaclass for the LayoutEnum."""

  @classmethod
  def __prepare__(mcls, name: str, bases: tuple[type], **kwargs) -> dict:
    """Returns the namespace."""
    bases = [b for b in bases if b.__class__ is not mcls]
    if not bases:
      return {}
    if len(bases) == 1 and bases[0] is object:
      return {}
    e = """EZnum classes do not support base classes, but received: <br>%s"""
    e2 = '<br><tab>'.join([str(b) for b in bases])
    raise TypeError(monoSpace(e % e2))

  @classmethod
  def _getAttrFactory(cls) -> Callable:
    """Returns the attribute factory."""

    def __getattr__(self, name: str) -> Any:
      """Returns the attribute."""
      try:
        return object.__getattribute__(self, name.lower())
      except AttributeError as attributeError:
        raise attributeError from AttributeError(name)

    return __getattr__

  @staticmethod
  def getBaseNameSpace() -> dict:
    """Returns the base namespace."""
    return {'name'                   : None, 'value': None,
            '__annotations__'        : {'name': str, 'value': int},
            '__members__'            : None,
            '__allow_instantiation__': True,
            '__iter_contents__'      : None, }

  @classmethod
  def _strFactory(mcls) -> Callable:
    """Returns the string representation of the class."""

    def __str__(self, ) -> str:
      """Returns the string representation of the class."""
      return self.name

    return __str__

  @classmethod
  def _reprFactory(mcls) -> Callable:
    """Returns the string representation of the class."""

    def __repr__(self, ) -> str:
      """Returns the string representation of the class."""
      clsName = self.__class__.__qualname__
      return '%s.%s' % (clsName, self.name,)

    return __repr__

  def __new__(mcls,
              name: str,
              bases: tuple[type],
              namespace: dict,
              **kwargs) -> type:
    enums = [k for (k, v) in namespace.items() if isinstance(v, auto)]
    enumSpace = {}
    coreSpace = dict(__name__=name,
                     __metaclass__=mcls,
                     __str__=mcls._strFactory(),
                     __repr__=mcls._reprFactory(),
                     __getattr__=mcls._getAttrFactory(), )
    baseSpace = mcls.getBaseNameSpace() | coreSpace
    cls = type.__new__(mcls, name, (), baseSpace)

    members = {}
    for (i, name) in enumerate(enums):
      enum = cls()
      setattr(enum, 'name', name)
      setattr(enum, 'value', i)
      setattr(cls, name, enum)
      members[name] = enum
    setattr(cls, '__members__', members)
    setattr(cls, '__allow_instantiation__', False)
    return cls

  def __init__(cls,
               name_: str,
               bases: tuple[type],
               namespace: dict,
               **kwargs) -> None:
    for (name, enum) in getattr(cls, '__members__').items():
      if name.lower() != name:
        setattr(cls, name.lower(), enum)
      if name.upper() != name:
        setattr(cls, name.upper(), enum)
    type.__init__(cls, name_, bases, namespace, **kwargs)

  def __iter__(cls) -> type:
    """Iterates through the members."""
    m = getattr(cls, '__members__', )
    setattr(cls, '__iter_contents__', [v for (k, v) in m.items()])
    return cls

  def __next__(cls) -> type:
    """Iterates through the members."""
    i = getattr(cls, '__iter_contents__', )
    try:
      value = i.pop(0)
      setattr(cls, '__iter_contents__', i)
      return value
    except IndexError:
      setattr(cls, '__iter_contents__', None)
      raise StopIteration

  def __str__(cls, ) -> str:
    """Returns the string representation of the class."""
    return cls.__name__

  def _lookUp(cls, name: str, **kwargs) -> Any:
    """Returns the enum member."""
    members = getattr(cls, '__members__')
    if name in members:
      return members[name]
    if name != name.lower():
      if name.lower() in members:
        return members[name.lower()]
    if name != name.upper():
      if name.upper() in members:
        return members[name.upper()]
    if kwargs.get('_recursion', True):
      val = kwargs.get('_recursion', 2)
      if not val:
        e = """Recursion limit reached!"""
        raise RecursionError(e)
      return cls._lookUp(name.lower(), )
    try:
      cls._lookUp(name.lower())
    except NameError as nameError:
      try:
        cls._lookUp(name.upper())
      except NameError as nameError3:
        raise nameError3
      e = """'%s' is not a member of the %s class."""
      raise NameError(e % (name, cls.__name__)) from nameError
    e = """'%s' is not a member of the %s class."""
    raise NameError(e % (name, cls.__name__))

  def __call__(cls, *args, **kwargs) -> Any:
    """Returns the enum member."""
    if not args and not kwargs:
      return type.__call__(cls, )
    if getattr(cls, '__allow_instantiation__', False):
      try:
        return cls._lookUp(args[0])
      except NameError as nameError:
        e = """Enum classes should not be manually instantiated!"""
        raise TypeError(e) from nameError
    return type.__call__(cls, *args, **kwargs)

  def __getitem__(cls, name: str) -> Any:
    """Returns the enum member."""
    if name in getattr(cls, '__members__', ).values():
      return name
    try:
      return cls._lookUp(name)
    except NameError as nameError:
      raise KeyError(name) from nameError

  def __contains__(cls, item: Any) -> bool:
    """Returns True if the name is a member."""
    for (key, val) in getattr(cls, '__members__', ).items():
      if item in [key, val]:
        return True
    return False


class _object(object):
  """_object provides the base class for the EZnum class."""

  def __init_subclass__(cls, **kwargs) -> None:
    """Initializes the subclass."""
    object.__init_subclass__()

  def __init__(self, *args, **kwargs) -> None:
    """We don't need keyword argument errors!"""
    object.__init__(self)


class EZnum(metaclass=EZnumMeta, _root=True):
  """Subclass EZnum instead of setting the metaclass. """

  def __init_subclass__(cls, **kwargs) -> None:
    """We don't need keyword argument errors!"""
    object.__init_subclass__()

  def __init__(self, *args, **kwargs) -> None:
    """We don't need keyword argument errors!"""
    object.__init__(self)

  def __get_attr__(self, name: str) -> Any:
    """Returns the attribute."""
    try:
      return object.__getattribute__(self, name.lower())
    except AttributeError as attributeError:
      raise attributeError from AttributeError(name)



"""The TextLabel class provides a text labels"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import QRect, QPoint
from PySide6.QtGui import QPainter, QPaintEvent, QFontMetrics
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from ezside.core import Center
from ezside.widgets import BaseWidget
from moreutils import StrField


class TextLabel(BaseWidget):
  """The TextLabel class provides a text labels"""

  __inner_text__ = None
  innerText = StrField('TextLabel')
  horizontalAlignment = StrField()
  verticalAlignment = StrField()

  def __init__(self, *args, **kwargs) -> None:
    BaseWidget.__init__(self, *args, **kwargs)
    textKeys = stringList("""msg, text, label, innerText""")
    for key in textKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if isinstance(val, str):
          self.__inner_text__ = val
          break
        else:
          e = typeMsg(key, val, str)
          raise TypeError(e)
    else:
      for arg in args:
        if isinstance(arg, str):
          self.__inner_text__ = arg
          break
      else:
        self.__inner_text__ = 'TextLabel'

  def getMetrics(self) -> QFontMetrics:
    """Returns the font metrics of the label."""
    return QFontMetrics(self.defaultFont)

  def getRect(self) -> QRect:
    """Returns the bounding rect of the label."""
    text = self.innerText
    return self.getMetrics().boundingRect(text)

  def setText(self, text: str) -> None:
    """Sets the text of the label."""
    self.innerText = text

  def getText(self, ) -> str:
    """Returns the text of the label."""
    return self.innerText

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the widget."""
    BaseWidget.initUi(self)
    self.defaultFont.setPointSize(12)
    rect = QFontMetrics(self.defaultFont).boundingRect(self.getText())
    self.setMinimumSize(rect.size())

  def alignRect(self, targetRect: QRect) -> QRect:
    """The rectangle returned by this method is the bounding rectangle
    required for the text placed in relation to the target rectangle."""
    hAlign = self.horizontalAlignment or 'center'
    vAlign = self.verticalAlignment or 'center'
    size = self.getMetrics().boundingRect(self.innerText).size()
    if hAlign == 'center':
      left = (targetRect.width() - size.width()) / 2
    elif hAlign == 'right':
      left = targetRect.width() - size.width()
    elif hAlign == 'left':
      left = 0
    else:
      e = """Invalid horizontal alignment: '%s'""" % (hAlign,)
      raise ValueError(e)
    if vAlign == 'center':
      top = (targetRect.height() - size.height()) / 2
    elif vAlign == 'top':
      top = 0
    elif vAlign == 'bottom':
      top = targetRect.height() - size.height()
    else:
      e = """Invalid vertical alignment: '%s'""" % (vAlign,)
      raise ValueError(e)
    left += targetRect.left()
    top += targetRect.top()
    leftTop = QPoint(left, top)
    return QRect(leftTop, size)

  def paintEvent(self, event: QPaintEvent) -> None:
    """The paintEvent method is called when the widget needs to be
    repainted."""
    painter = QPainter()
    painter.begin(self)
    viewRect = painter.viewport()
    textSize = self.getRect().size()
    # # # # # # # # # # # # # # # # #
    # Painting the fill
    painter.setPen(self.emptyLine)
    painter.setBrush(self.solidBrush)
    painter.drawRect(viewRect)
    # # # # # # # # # # # # # # # # #
    # Painting the border
    painter.setPen(self.solidLine)
    painter.setBrush(self.emptyBrush)
    painter.drawRect(viewRect)
    # # # # # # # # # # # # # # # # #
    # Painting the text

    painter.setPen(self.fontLine)
    painter.setFont(self.defaultFont)
    flags = Center
    text = self.getText()
    painter.drawText(viewRect, flags, text)
    painter.end()



"""The 'qt_names' package provides some shorter names for commonly used Qt
enum values and classes."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QFont, QTextOption
from PySide6.QtWidgets import QSizePolicy

SolidFill = Qt.BrushStyle.SolidPattern
BlankFill = Qt.BrushStyle.NoBrush

SolidLine = Qt.PenStyle.SolidLine
DashLine = Qt.PenStyle.DashLine
DotLine = Qt.PenStyle.DotLine
DashDot = Qt.PenStyle.DashDotLine
BlankLine = Qt.PenStyle.NoPen

FlatCap = Qt.PenCapStyle.FlatCap
SquareCap = Qt.PenCapStyle.SquareCap
RoundCap = Qt.PenCapStyle.RoundCap

MiterJoin = Qt.PenJoinStyle.MiterJoin
BevelJoin = Qt.PenJoinStyle.BevelJoin
RoundJoin = Qt.PenJoinStyle.RoundJoin
SvgMiterJoin = Qt.PenJoinStyle.SvgMiterJoin

Normal = QFont.Weight.Normal
Bold = QFont.Weight.Bold
DemiBold = QFont.Weight.DemiBold

WrapMode = QTextOption.WrapMode
NoWrap = QTextOption.WrapMode.NoWrap
WordWrap = QTextOption.WrapMode.WordWrap

AlignFlag = Qt.AlignmentFlag
AlignLeft = Qt.AlignmentFlag.AlignLeft
AlignRight = Qt.AlignmentFlag.AlignRight
AlignHCenter = Qt.AlignmentFlag.AlignHCenter
AlignVCenter = Qt.AlignmentFlag.AlignVCenter
AlignCenter = Qt.AlignmentFlag.AlignCenter
Center = Qt.AlignmentFlag.AlignCenter
AlignTop = Qt.AlignmentFlag.AlignTop
AlignBottom = Qt.AlignmentFlag.AlignBottom

Expand = QSizePolicy.Policy.MinimumExpanding
Tight = QSizePolicy.Policy.Maximum
Fixed = QSizePolicy.Policy.Fixed

TimerType = Qt.TimerType
Precise = Qt.TimerType.PreciseTimer
Coarse = Qt.TimerType.CoarseTimer
VeryCoarse = Qt.TimerType.VeryCoarseTimer



"""Provides defaults settings in the Default class"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations


class Default:
  """Provides defaults settings in the Default class"""

  anyTimer = 25
  paintTimer = 20
  noiseTimer = 10
  numPoints = 256

  sliderMin = 0
  sliderMax = 100
  sliderSingleStep = 1
  sliderPageStep = 10
  sliderTickInterval = 10

  numSliders = 8

  chartViewWidth = 400
  chartViewHeight = 300

  layoutPreference = 0  # 0: vertical, 1: horizontal

  bannerMargin = 32
  sliderHeight = 192
  sliderWidth = 32

  fontFamily = 'Montserrat'



"""CornerPanel provides corners together with horizontal and vertical
panels."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import QSize

from ezside.widgets import VerticalPanel
from settings import Default


class CornerPanel(VerticalPanel):
  """CornerPanel provides corners together with horizontal and vertical
  panels."""

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    m = Default.bannerMargin
    self.setFixedSize(QSize(m, m))



"""Brush class for the EZQt library."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QColor


class Brush(QBrush):
  """Brush class for the EZQt library."""

  def __init__(self, *args, **kwargs) -> None:
    QBrush.__init__(self, )
    intArgs = [arg for arg in args if isinstance(arg, int)]
    colorArgs = [arg for arg in args if isinstance(arg, QColor)]
    styleArgs = [arg for arg in args if isinstance(arg, Qt.BrushStyle)]
    color = None
    style = None
    if colorArgs:
      color = colorArgs[0]
    elif len(intArgs) in [3, 4]:
      color = QColor(*intArgs)
    else:
      color = QColor(255, 255, 255, 255)
    if styleArgs:
      style = styleArgs[0]
    else:
      style = Qt.BrushStyle.SolidPattern
    self.setColor(color)
    self.setStyle(style)



"""BorderRect provides a bordered rectangle with no fill."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QPaintEvent, QPainter

from ezside.core import emptyBrush
from ezside.widgets import BaseWidget


class BorderRect(BaseWidget):
  """BorderRect provides a bordered rectangle with no fill."""

  def paintEvent(self, event: QPaintEvent) -> None:
    """The paintEvent method is called when the widget needs to be
    repainted."""
    painter = QPainter()
    painter.begin(self)
    painter.setPen(self.solidLine)
    painter.setBrush(emptyBrush())
    painter.drawRect(self.rect())
    painter.end()



"""MenuBar provides the menu bar for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMenuBar, QMainWindow
from icecream import ic

from attribox import AttriBox, this
from ezside.windows.menus import FilesMenu, EditMenu, HelpMenu, DebugMenu

ic.configureOutput(includeContext=True, )


class MenuBar(QMenuBar):
  """MenuBar provides the menu bar for the main application window."""

  files = AttriBox[FilesMenu](this)
  edit = AttriBox[EditMenu](this)
  help = AttriBox[HelpMenu](this)
  debug = AttriBox[DebugMenu](this)

  def __init__(self, parent: QMainWindow) -> None:
    QMenuBar.__init__(self, parent)

  def initUi(self, ) -> None:
    """Initializes the user interface for the menu bar."""
    self.files.initUi()
    self.addMenu(self.files)
    self.edit.initUi()
    self.addMenu(self.edit)
    self.help.initUi()
    self.addMenu(self.help)
    self.debug.initUi()
    self.addMenu(self.debug)



"""The 'ezside.core' package provides a limited selection from the Qt
namespace in much shorter named versions. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._qt_names import *
from ._colors import *

from ._pen import Pen
from ._brush import Brush
from ._parse_parent import parseParent
from ._font_family import Font

from ._color_factory import parseColor
from ._pen_factory import parsePen, stylePen, emptyPen, solidPen
from ._brush_factory import parseBrush, emptyBrush, solidBrush



"""BoxLayout provides single row or column layouts."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from PySide6.QtWidgets import QLayout

from ezside.widgets import BaseWidget


class BoxLayout(BaseWidget):
  """BoxLayout provides single row or column layouts."""
  __base_layout__ = None

  def __init__(self, *args, **kwargs) -> None:
    BaseWidget.__init__(self, *args, **kwargs)
    self.__child_widgets__ = []

  @abstractmethod
  def createBaseLayout(self) -> None:
    """Creates the base layout."""

  def getBaseLayout(self, **kwargs) -> QLayout:
    """Returns the base layout."""
    if self.__base_layout__ is None:
      if kwargs.get('_recursion', None):
        raise RecursionError
      self.createBaseLayout()
      return self.getBaseLayout(_recursion=True)
    return self.__base_layout__

  def setBaseLayout(self, layout: QLayout) -> None:
    """Sets the base layout."""
    self.__base_layout__ = layout

  def addWidget(self, *args) -> None:
    """Adds a widget to the layout. """
    for arg in args:
      if isinstance(arg, BaseWidget):
        widget = arg
        break
    else:
      raise ValueError('Received no widget!')
    self.__child_widgets__.append(widget)
    widget.initUi()
    widget.connectActions()
    self.getBaseLayout().addWidget(widget)

  def reset(self) -> None:
    """Rebuilds the layout."""
    setattr(self, '__base_layout__', None)
    self.createBaseLayout()
    for widget in self.__child_widgets__:
      self.addWidget(widget)
    self.setLayout(self.getBaseLayout())



"""Action subclasses QAction and customizes the instantiation process."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QAction, QKeySequence
from icecream import ic
from vistutils.waitaminute import typeMsg

from ezside.windows.menus.shortcuts import getShortcut
from settings import Icons

ic.configureOutput(includeContext=True, )


class Action(QAction):
  """Action subclasses QAction and customizes the instantiation process."""

  def __init__(self, *args, **kwargs) -> None:
    if isinstance(args[0], str):
      name = args[0]
      QAction.__init__(self, name)
      self._name = name
    elif isinstance(args[0], QAction):
      self._name = args[0].text() or 'LOL'
    else:
      e = typeMsg('action', args[0], QAction)
      raise TypeError(e)
    self.__icon_path__ = None

  def setIconPath(self, iconPath: str) -> None:
    """Sets the icon for the action."""
    self.__icon_path__ = iconPath

  def setupAction(self, ) -> None:
    """Sets up the action."""
    shortcut = getShortcut(self._name)
    if isinstance(shortcut, QKeySequence):
      self.setShortcut(shortcut)
    self.setupIcon()

  def setupIcon(self) -> None:
    """Sets up the icon for the action."""
    self.setIcon(Icons.load(self._name))

  def __str__(self, ) -> str:
    """Returns the name of the action."""
    name = self._name
    shortcut = self.shortcut().toString()
    menu = self.menu().title()
    if shortcut:
      return '%s | %s %s' % (menu, name, shortcut)
    return '%s | %s' % (menu, name)

  def __repr__(self, ) -> str:
    """Returns the name of the action."""
    return self._name



"""DataView collects the functionality from QCharts to show data in a plot
that is continuously updated."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCharts import QChart, QChartView
from PySide6.QtCore import Slot
from icecream import ic
from attribox import AttriBox

from ezside.widgets import ScatterSeries, LineSeries
from settings import Default

if TYPE_CHECKING:
  pass

ic.configureOutput(includeContext=True, )


class DataView(QChartView):
  """DataView collects the functionality from QCharts to show data in a plot
  that is continuously updated."""

  innerChart = AttriBox[QChart]()
  lineVals = AttriBox[LineSeries]()
  scatter = AttriBox[ScatterSeries]()

  def initUi(self) -> None:
    """Sets up the view"""
    self.setMinimumSize(Default.chartViewWidth, Default.chartViewHeight)
    self.initScatter()

  def initScatter(self, ) -> None:
    """Initializes the scatter series."""
    self.setChart(self.innerChart)
    self.innerChart.removeAllSeries()
    self.innerChart.addSeries(self.scatter)
    self.innerChart.createDefaultAxes()
    self.update()

  def initLine(self) -> None:
    """Initializes the line series."""
    self.setChart(self.innerChart)
    self.innerChart.removeAllSeries()
    self.innerChart.addSeries(self.lineVals)
    self.innerChart.createDefaultAxes()
    self.update()

  @Slot(float)
  def append(self, value: float) -> None:
    """Appends a value to the series."""
    self.scatter.appendValue(value)
    self.lineVals.appendValue(value)

  @Slot()
  def refresh(self) -> None:
    """Refreshes the data."""
    self.scatter.updateValues()
    self.lineVals.updateValues()
    self.update()



"""ClientInfo shows static information about the current client"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QVBoxLayout
from attribox import AttriBox

from ezside.core import Tight
from ezside.widgets import BaseWidget, TextLabel


class ClientInfo(BaseWidget):
  """ClientInfo shows static information about the current client"""

  uri = AttriBox[TextLabel]('URI')
  robotId = AttriBox[TextLabel]('Robot ID')
  customerId = AttriBox[TextLabel]('Customer ID')
  baseLayout = AttriBox[QVBoxLayout]()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(128, 128)
    self.uri.initUi()
    self.uri.setSizePolicy(Tight, Tight)
    self.baseLayout.addWidget(self.uri)
    self.robotId.initUi()
    self.robotId.setSizePolicy(Tight, Tight)
    self.baseLayout.addWidget(self.robotId)
    self.customerId.initUi()
    self.customerId.setSizePolicy(Tight, Tight)
    self.baseLayout.addWidget(self.customerId)
    self.setLayout(self.baseLayout)



"""The 'moreutils' package provides a collection of utility classes and
functions. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._int_field import IntField
from ._str_field import StrField



"""The 'ezside.windows' package provides convenient function for working
with the main application window when developing in pyside6."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._base_window import BaseWindow
from ._layout_window import LayoutWindow
from ._main_window import MainWindow
from ._test_window import TestWindow



"""VerticalSlider provides a vertical slider."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Callable, Self

from PySide6.QtCore import Qt, QSize
from PySide6.QtWidgets import QSlider
from vistutils.parse import maybe
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from ezside.core import Expand
from settings import Default


class VerticalSlider(QSlider):
  """VerticalSlider provides a vertical slider."""

  __value_map__ = None

  @classmethod
  def defaultInstance(cls) -> VerticalSlider:
    """Return the default instance."""
    return cls()

  @classmethod
  def clone(cls, slider: VerticalSlider = None) -> VerticalSlider:
    """Clone the given slider."""
    if slider is None:
      return cls.defaultInstance()
    if isinstance(slider, (cls, QSlider)):
      return cls(slider.minimum(), slider.maximum(), slider.singleStep(),
                 slider.pageStep(), slider.value())
    e = typeMsg('slider', slider, cls)
    raise TypeError(e)

  def __init__(self, *args, **kwargs) -> None:
    minVal, maxVal, singleStep, pageStep, initVal = [None] * 5
    minValKeys = stringList("""min, minVal, minimum, minVal""")
    maxValKeys = stringList("""max, maxVal, maximum, maxVal""")
    singleStepKeys = stringList("""singleStep, step, single, stepSize""")
    pageStepKeys = stringList("""pageStep, page, pageStepSize""")
    initValKeys = stringList("""initVal, value, init, val""")
    Keys = [minValKeys, maxValKeys, singleStepKeys, pageStepKeys,
            initValKeys]
    for keys in Keys:
      for key in keys:
        if key in kwargs:
          val = kwargs.get(key)
          if key in minValKeys:
            if isinstance(val, int):
              minVal = val
              break
          elif key in maxValKeys:
            if isinstance(val, int):
              maxVal = val
              break
          elif key in singleStepKeys:
            if isinstance(val, int):
              singleStep = val
              break
          elif key in pageStepKeys:
            if isinstance(val, int):
              pageStep = val
              break
          elif key in initValKeys:
            if isinstance(val, int):
              initVal = val
              break
    else:
      for arg in args:
        if isinstance(arg, int):
          if minVal is None:
            minVal = arg
          elif maxVal is None:
            maxVal = arg
          elif singleStep is None:
            singleStep = arg
          elif pageStep is None:
            pageStep = arg
          elif initVal is None:
            initVal = arg
    minVal = maybe(minVal, Default.sliderMin)
    maxVal = maybe(maxVal, Default.sliderMax)
    singleStep = maybe(singleStep, Default.sliderSingleStep)
    pageStep = maybe(pageStep, Default.sliderPageStep)
    QSlider.__init__(self, Qt.Orientation.Vertical)
    self.setRange(minVal, maxVal)
    self.setSingleStep(singleStep)
    self.setPageStep(pageStep)
    self.setTickInterval(pageStep)
    self.setValue(maybe(initVal, minVal))
    self.applyStyleSheet()

  def applyStyleSheet(self) -> None:
    """
    Applies the stylesheet to the slider to make the handle larger
    and potentially apply other custom styles.
    """
    self.setStyleSheet("""
      QSlider::handle:horizontal {
        background: #5c5c5c;
        border: 1px solid #5c5c5c;
        width: 20px;
        margin: -10px 0; /* expand outside the container */
      }

      QSlider::handle:vertical {
        background: #5c5c5c;
        border: 1px solid #5c5c5c;
        height: 20px;
        margin: 0 -10px; /* expand outside the container */
      }
    """)

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setSizePolicy(Expand, Expand)
    self.setMinimumSize(QSize(Default.sliderWidth, Default.sliderHeight))

  def __abs__(self) -> int:
    """Return the absolute value of the slider."""
    return self.maximum() - self.minimum()

  def __float__(self) -> float:
    """Return the float value of the slider."""
    baseValue = float(self.value()) / abs(self)
    if self.__value_map__ is None:
      return baseValue
    if callable(self.__value_map__):
      return self.__value_map__(baseValue)
    try:
      f = float(self.__value_map__)
    except Exception as exception:
      e = typeMsg('value_map', self.__value_map__, Callable)
      raise TypeError(e) from exception
    return f * baseValue

  def __int__(self) -> int:
    """Return the integer value of the slider."""
    return self.value()

  def __neg__(self) -> int:
    """Return the negated value of the slider."""
    return self.maximum() - self.value()

  def __pos__(self) -> int:
    """Return the positive value of the slider."""
    return self.value() - self.minimum()

  def __imul__(self, value: float) -> Self:
    """Set value map to float."""
    if not isinstance(value, (int, float)):
      e = typeMsg('value', value, float)
      raise TypeError(e)
    self.__value_map__ = value
    return self

  def __imatmul__(self, callMeMaybe: Callable) -> Self:
    """Set value map to callable"""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__value_map__ = callMeMaybe
    return self



"""Horizontal layout widget."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QHBoxLayout

from ezside.widgets import BoxLayout


class HLayout(BoxLayout):
  """Horizontal layout widget."""

  def createBaseLayout(self) -> None:
    """Creates the base layout."""
    self.setBaseLayout(QHBoxLayout())



"""KeyMap provides named instances of QKeySequence for use in the main
application."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from warnings import warn

from PySide6.QtGui import QKeySequence
from icecream import ic
from vistutils.text import monoSpace

ic.configureOutput(includeContext=True, )


def _getKeyboardShortcuts() -> dict:
  """
  Returns a dictionary of common keyboard shortcuts
  categorized by Files, Edit, View, and Help actions.
  """
  return {
    "new"       : "CTRL+N",
    "open"      : "CTRL+O",
    "save"      : "CTRL+S",
    "save_as"   : "CTRL+SHIFT+S",
    "close"     : "CTRL+W",
    "undo"      : "CTRL+Z",
    "redo"      : "CTRL+Y",
    "cut"       : "CTRL+X",
    "copy"      : "CTRL+C",
    "paste"     : "CTRL+V",
    "select_all": "CTRL+A",
    "about_qt"  : "F12",
    "exit"      : "ALT+F4",
  }


def getShortcut(name: str) -> QKeySequence:
  """
  Returns a QKeySequence for the given name.
  """
  name = name.lower().strip()
  name = name.replace(' ', '_')
  if 'debug' in name:
    for char in name:
      if char.isnumeric():
        return QKeySequence('F%s' % char)
  base = _getKeyboardShortcuts()
  shortcut = base.get(name, '')
  sequence = QKeySequence.fromString(shortcut)
  if not sequence:
    warn(f'No keyboard shortcut found for {monoSpace(name)}')
    return QKeySequence()
  return sequence



"""Font wraps the QFont customizing some controls"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QFont, QFontMetrics
from PySide6.QtGui import QFontDatabase


class Font(QFont):
  """FontFamily wraps the QFont family property"""

  @staticmethod
  def resolveFontFamily(family: str) -> str:
    """Resolves the font family name"""
    font_families = QFontDatabase().families()

    if family in font_families:
      return family
    elif family == 'monospace':
      return 'Courier'
    elif family == 'serif':
      return 'Times'
    elif family == 'sans-serif':
      return 'Helvetica'
    else:
      return 'Helvetica'  # default font family

  def __init__(self, *args, **kwargs) -> None:
    family = None
    font = None
    pointSize = None
    for arg in args:
      if isinstance(arg, str) and family is None:
        family = self.resolveFontFamily(arg) or None
      if isinstance(arg, QFont) and font is None:
        font = arg
      if isinstance(arg, int) and pointSize is None:
        pointSize = arg
    QFont.__init__(self, )
    self.setFamily(family or 'Helvetica')
    self.setPointSize(pointSize or 16)
    if font:
      self.setBold(font.bold())
      self.setItalic(font.italic())
      self.setUnderline(font.underline())
      self.setStrikeOut(font.strikeOut())
      self.setWeight(font.weight())
      self.setCapitalization(font.capitalization())
      self.setOverline(font.overline())
      self.setPixelSize(font.pixelSize())
      self.setStretch(font.stretch())
      self.setWordSpacing(font.wordSpacing())
      self.setStyle(font.style())
      self.setFixedPitch(font.fixedPitch())

  def metrics(self, ) -> QFontMetrics:
    """Returns the font metrics"""
    return QFontMetrics(self)



"""The DataEcho class provides a way to echo data from one object to
another."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time

import numpy as np
from attribox import AttriBox
from icecream import ic
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

ic.configureOutput(includeContext=True, )


class DataRoll:
  """The DataEcho class provides a way to echo data from one object to
  another."""

  def __init__(self, *args, **kwargs) -> None:
    self.__zero_index__ = 0
    self.__zero_time__ = time.time()
    self.__num_points__ = None
    if args:
      attriName = getattr(AttriBox, '__qualname__', None)
      if attriName is not None:
        if isinstance(attriName, str):
          boxName = '__boxes_%s__' % attriName
          if getattr(args[0].__class__, boxName, None) is not None:
            numPoints = getattr(args[0], 'getNumPoints', None)
            if numPoints is not None:
              if callable(numPoints):
                self.__num_points__ = numPoints()
    if self.__num_points__ is None:
      numKeys = stringList("""n, num, num_points, numPoints""")
      for key in numKeys:
        if key in kwargs:
          val = kwargs.get(key)
          if isinstance(val, int):
            self.__num_points__ = val
            break
          else:
            e = typeMsg(key, val, int)
            raise TypeError(e)
      else:
        for arg in args:
          if isinstance(arg, int):
            self.__num_points__ = arg
            break
        else:
          self.__num_points__ = 256
    self.__inner_array__ = None
    self.createArray()

  def createArray(self, ) -> None:
    """Creates the array."""
    n = self.__num_points__
    t = np.linspace(0, n - 1, n, dtype=np.complex64) / (n - 1)
    y = np.zeros(n, dtype=np.complex64)
    self.__inner_array__ = t + y * 1j

  def append(self, value: float = None) -> None:
    """Append a value to the DataEcho object."""
    if not value == value:
      value = 0
    entry = time.time() + value * 1j - self.__zero_time__
    self.__inner_array__[self.__zero_index__] = entry
    self.__zero_index__ = (self.__zero_index__ + 1) % len(self)

  def __len__(self, ) -> int:
    """Returns the length of the array."""
    return len(self.__inner_array__)

  def snapShot(self, ) -> np.ndarray:
    """Returns the array."""
    out = self.__inner_array__.copy()
    out = np.roll(out, -self.__zero_index__)
    zeroTime = out.real.min()
    return out - zeroTime



"""Timer wraps the QTimer class."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import QTimer
from vistutils.parse import maybe
from vistutils.text import stringList

from ezside.core import Precise, TimerType
from settings import Default


class Timer(QTimer):
  """Timer wraps the QTimer class."""

  def __init__(self, *args, **kwargs) -> None:
    interval = None
    timerType = None
    singleShot = None
    intervalKeys = stringList("""interval, time, epoch, period""")
    typeKeys = stringList("""timerType, type, mode""")
    singleShotKeys = stringList("""singleShot, oneShot, once""")
    Keys = [intervalKeys, typeKeys, singleShotKeys]
    for keys in Keys:
      for key in keys:
        if key in kwargs:
          val = kwargs.get(key)
          if key in intervalKeys:
            if isinstance(val, int):
              interval = val
              break
          elif key in typeKeys:
            if isinstance(val, int):
              timerType = val
              break
          elif key in singleShotKeys:
            if isinstance(val, bool):
              singleShot = val
              break
    else:
      for arg in args:
        if isinstance(arg, int):
          interval = arg
        if isinstance(arg, bool):
          singleShot = arg
        if isinstance(arg, TimerType):
          timerType = arg
      interval = maybe(interval, Default.anyTimer)
      timerType = maybe(timerType, Precise)
      singleShot = maybe(singleShot, False)
    QTimer.__init__(self, )
    self.setInterval(interval)
    self.setTimerType(timerType)
    self.setSingleShot(singleShot)



"""The 'eznum' package provides enum types."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._auto import auto
from ._eznum_meta import EZnumMeta, EZnum



"""HelpMenu provides the help menu for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow

from attribox import AttriBox
from ezside.windows.menus import AbstractMenu, Action


class HelpMenu(AbstractMenu):
  """HelpMenu provides the help menu for the main application window."""

  about_qt = AttriBox[Action]('about_qt')
  about_conda = AttriBox[Action]('about_conda')
  about_python = AttriBox[Action]('about_python')

  def __init__(self, parent: QMainWindow) -> None:
    """HelpMenu provides the help menu for the main application window."""
    AbstractMenu.__init__(self, parent)
    self.setTitle('Help')

  def appendActions(self) -> None:
    """Appends the actions without setting them up"""
    self.addAction(self.about_qt)
    self.addAction(self.about_conda)
    self.addAction(self.about_python)

  def setupActions(self, ) -> None:
    """Sets up the actions for the help menu."""
    self.about_qt.setupAction()
    self.about_conda.setupAction()
    self.about_python.setupAction()



"""The 'ezside' package provides convenient implementation of features from
the pyside6 framework."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ezside.windows.menus import Action
from ezside.windows.menus import AbstractMenu
from ezside.windows.menus import FilesMenu
from ezside.windows.menus import EditMenu
from ezside.windows.menus import HelpMenu
from ezside.windows.menus import DebugMenu
from ezside.windows.bars import MenuBar
from ezside.windows.bars import StatusBar
from ezside.windows import BaseWindow
from ezside.windows import LayoutWindow
from ezside.windows import MainWindow



"""The parseParent function parses positional arguments and returns the
first instance of QWidget encountered if any."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Optional

from PySide6.QtWidgets import QWidget
from vistutils.text import stringList


def parseParent(*args, **kwargs) -> Optional[QWidget]:
  """The parseParent function parses positional arguments and returns the
  first instance of QWidget encountered if any."""
  parentKeys = stringList("""parent, main, window""")
  for key in parentKeys:
    if key in kwargs:
      val = kwargs.get(key)
      if isinstance(val, QWidget):
        return val
  for arg in args:
    if isinstance(arg, QWidget):
      return arg



"""The 'ezside.windows.bars' package provides convenient implementations of
the status bar and menu bar for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._menu_bar import MenuBar
from ._status_bar import StatusBar



"""RadioButton provides a button that can be turned on by a mouse click,
but which require input from somewhere else to reactivate."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen

from ezside.widgets import BaseWidget


class RadioButton(BaseWidget):
  """AnimatedToggleButton provides a button that can be turned on by a
  mouse"""

  __current_state__ = False



"""Spinbox wraps QSpinBox"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Signal
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QDoubleSpinBox, QHBoxLayout, QGridLayout
from attribox import AttriBox

from ezside.widgets import BaseWidget, TextLabel, HSpacer, VSpacer
from settings import Default


class _Spinbox(QDoubleSpinBox):
  """Spinbox wraps QSpinBox"""

  def __init__(self, *args, **kwargs) -> None:
    minVal, maxVal = [*args, None, None][:2]
    QDoubleSpinBox.__init__(self)
    self.setRange(minVal, maxVal)
    span = maxVal - minVal
    self.setSingleStep(span / 100)
    self.setSuffix(' ~')
    self.setPrefix('~ ')

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the widget."""
    self.setMinimumHeight(32)


class SpinBox(BaseWidget):
  """Wrapper showing the Spinbox widget."""

  __label_font__ = None

  valueChanged = Signal(int)
  innerBox = AttriBox[_Spinbox](-5, 5, )
  label = AttriBox[TextLabel]()
  baseLayout = AttriBox[QGridLayout]()
  vSpacer = AttriBox[VSpacer]()
  hSpacer = AttriBox[HSpacer]()

  def __init__(self, name: str = None, *args, **kwargs) -> None:
    """The __init__ method initializes the SpinBox widget."""
    BaseWidget.__init__(self, *args, **kwargs)
    self._name = name
    self.label.innerText = name
    self.initUi()
    self.connectActions()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the widget."""
    self.innerBox.initUi()
    self.baseLayout.addWidget(self.innerBox, 0, 0, )
    self.label.defaultFont.setPointSize(12)
    self.label.initUi()
    self.baseLayout.addWidget(self.label, 1, 0, )
    # self.baseLayout.addWidget(self.hSpacer, 0, 1, 2, 1)

    self.setLayout(self.baseLayout)

  def connectActions(self) -> None:
    """The connectActions method connects the widget actions."""
    self.innerBox.valueChanged.connect(self.valueChanged.emit)

  def setValue(self, value: float) -> None:
    """The setValue method sets the value of the widget."""
    self.innerBox.setValue(value)

  def value(self) -> float:
    """The value method returns the value of the widget."""
    return self.innerBox.value()



"""The FilledRect subclass of BaseWidget provides a filled rectangle."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QColor, QPaintEvent, QPainter

from attribox import AttriBox

from ezside.core import parseColor, emptyPen

from ezside.widgets import BaseWidget


class FilledRect(BaseWidget):
  """The FilledRect subclass of BaseWidget provides a filled rectangle."""

  fillColor = AttriBox[QColor]()

  def __init__(self, *args, **kwargs) -> None:
    """The __init__ method initializes the FilledRect widget."""
    BaseWidget.__init__(self, *args, **kwargs)
    self.solidBrush.setColor(parseColor(self.fillColor))

  def paintEvent(self, event: QPaintEvent) -> None:
    """The paintEvent method is called when the widget needs to be
    repainted."""
    painter = QPainter()
    painter.begin(self)
    painter.setPen(emptyPen())
    painter.setBrush(self.solidBrush)
    painter.drawRoundedRect(self.rect(), 8, 8)
    painter.end()



"""MainWindow subclasses the LayoutWindow and provides the main
application business logic."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QMainWindow
from icecream import ic
from attribox import AttriBox

from ezside.core import Precise
from ezside.widgets import Timer
from ezside.windows import LayoutWindow
from settings import Default

ic.configureOutput(includeContext=True, )


class MainWindow(LayoutWindow):
  """MainWindow subclasses the LayoutWindow and provides the main
  application business logic."""

  paintTimer = AttriBox[Timer](Default.paintTimer, Precise, False)

  def __init__(self, *args, **kwargs) -> None:
    LayoutWindow.__init__(self, *args, **kwargs)
    for arg in args:
      if isinstance(arg, str):
        self.titleBanner.setText(arg)
        break
    else:
      self.titleBanner.setText('EZSide')

  def connectActions(self, ) -> None:
    """Connects the actions to the slots."""
    LayoutWindow.connectActions(self)
    self.whiteNoise.noise.connect(self.dataWidget.dataView.append)
    self.dataWidget.start.connect(self.startHandle)
    self.dataWidget.stop.connect(self.stopHandle)
    self.paintTimer.timeout.connect(self.dataWidget.refresh)
    self.plotEnvelope.newValues.connect(self.applyView)
    self.paintTimer.start()

  @Slot()
  def stopHandle(self, ) -> None:
    """Handles the stop button"""

  @Slot()
  def startHandle(self, ) -> None:
    """Handles the start button"""

  def applyView(self, ) -> None:
    """Applies the view to the data widget."""
    values = self.plotEnvelope.getValues()

    self.dataWidget.dataView.chart().axisX().setMin(values.get('minH'))
    self.dataWidget.dataView.chart().axisY().setMin(values.get('minV'))
    self.dataWidget.dataView.chart().axisX().setMax(values.get('maxH'))
    self.dataWidget.dataView.chart().axisY().setMax(values.get('maxV'))

    self.dataWidget.dataView.refresh()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    LayoutWindow.initUi(self, )

  def show(self, ) -> None:
    """Shows the window."""
    self.initUi()
    self.connectActions()
    QMainWindow.show(self)



"""PlotEnvelope widget module."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QGridLayout
from attribox import AttriBox

from ezside.widgets import BaseWidget, SpinBox, PushButton


class PlotEnvelope(BaseWidget):
  """PlotEnvelope widget."""

  __min_horizontal_value__ = 0
  __max_horizontal_value__ = 1
  __min_vertical_value__ = 0
  __max_vertical_value__ = 1

  valueChanged = Signal()
  newValues = Signal()
  cancelled = Signal()

  basePlot = AttriBox[QGridLayout]()
  minHorizontal = AttriBox[SpinBox]('Min. Horizontal')
  maxHorizontal = AttriBox[SpinBox]('Max. Horizontal')
  minVertical = AttriBox[SpinBox]('Min. Vertical')
  maxVertical = AttriBox[SpinBox]('Max. Vertical')
  apply = AttriBox[PushButton]('Apply')
  cancel = AttriBox[PushButton]('Cancel')

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the widget."""
    self.setMinimumSize(96, 96)
    self.minHorizontal.initUi()
    self.minHorizontal.setValue(-1)
    self.basePlot.addWidget(self.apply, 0, 0)
    self.maxHorizontal.initUi()
    self.maxHorizontal.setValue(1)
    self.basePlot.addWidget(self.cancel, 1, 0)
    self.minHorizontal.initUi()
    self.minHorizontal.setValue(-1)
    self.basePlot.addWidget(self.minHorizontal, 0, 1)
    self.maxHorizontal.initUi()
    self.maxHorizontal.setValue(1)
    self.basePlot.addWidget(self.maxHorizontal, 1, 1)
    self.minVertical.initUi()
    self.minVertical.setValue(-1)
    self.basePlot.addWidget(self.minVertical, 0, 2)
    self.maxVertical.initUi()
    self.maxVertical.setValue(1)
    self.basePlot.addWidget(self.maxVertical, 1, 2)
    self.setLayout(self.basePlot)
    self.connectActions()

  def connectActions(self) -> None:
    """The connectActions method connects the widget actions."""
    self.apply.clicked.connect(self.newValues.emit)
    self.cancel.clicked.connect(self.cancelled.emit)
    self.minHorizontal.valueChanged.connect(self.handleValueChanged)
    self.maxHorizontal.valueChanged.connect(self.handleValueChanged)
    self.minVertical.valueChanged.connect(self.handleValueChanged)
    self.maxVertical.valueChanged.connect(self.handleValueChanged)

  def handleApply(self) -> None:
    """Handle the apply button."""
    self.__min_horizontal_value__ = self.minHorizontal.value()
    self.__max_horizontal_value__ = self.maxHorizontal.value()
    self.__min_vertical_value__ = self.minVertical.value()
    self.__max_vertical_value__ = self.maxVertical.value()
    self.apply.setEnabled(False)
    self.newValues.emit()

  def handleCancel(self) -> None:
    """Handle the cancel button."""
    self.minHorizontal.setValue(self.__min_horizontal_value__)
    self.maxHorizontal.setValue(self.__max_horizontal_value__)
    self.minVertical.setValue(self.__min_vertical_value__)
    self.maxVertical.setValue(self.__max_vertical_value__)
    self.cancel.setEnabled(False)
    self.cancelled.emit()

  def handleValueChanged(self) -> None:
    """Handle the value changed signal."""
    self.apply.setEnabled(True)
    self.cancel.setEnabled(True)
    self.valueChanged.emit()

  def getValues(self) -> dict[str, float]:
    """The getValueDict method returns the values of the widget."""
    return {'minH': self.minHorizontal.value(),
            'maxH': self.maxHorizontal.value(),
            'minV': self.minVertical.value(),
            'maxV': self.maxVertical.value()}



"""Layout provides the request type of QLayout, either vertical,
horizontal or gridded. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtWidgets import QGridLayout, QWidget
from attribox import AttriBox
from icecream import ic
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from eznum import EZnum, auto
from ezside.widgets import BaseWidget


class LayoutEnum(EZnum):
  """LayoutEnum provides the request type of QLayout, either vertical,
  horizontal or gridded. """
  VERTICAL = auto()
  HORIZONTAL = auto()
  GRID = auto()


class InnerLayout(QGridLayout):
  """Layout wrapper"""

  __col_count__ = 0
  __row_count__ = 0
  __total_count__ = 0

  def __init__(self, layoutType: Any = None) -> None:
    """Initializes the Layout instance. """
    QGridLayout.__init__(self, )
    self.__layout_mode__ = None
    if isinstance(layoutType, LayoutEnum):
      self.__layout_mode__ = layoutType
    elif isinstance(layoutType, str):
      if layoutType in LayoutEnum:
        self.__layout_mode__ = LayoutEnum[layoutType]
    else:
      self.__layout_mode__ = LayoutEnum['grid']

  def addWidget(self, *args) -> None:
    """Adds a widget to the layout. """
    if self.__layout_mode__ == LayoutEnum.HORIZONTAL:
      self.addHorizontalWidget(*args)
    elif self.__layout_mode__ == LayoutEnum.VERTICAL:
      self.addVerticalWidget(*args)
    elif self.__layout_mode__ == LayoutEnum.GRID:
      self.addGridWidget(*args)

  def addHorizontalWidget(self, *args) -> None:
    """Adds a widget to the layout. """
    widget = [*[a for a in args if isinstance(a, QWidget)], None][0]
    if isinstance(widget, QWidget):
      QGridLayout.addWidget(self, widget, 0, self.__total_count__, )
      self.__total_count__ += 1
      self.__col_count__ += 1

  def addVerticalWidget(self, *args) -> None:
    """Adds a widget to the layout. """
    widget = [*[a for a in args if isinstance(a, QWidget)], None][0]
    if isinstance(widget, QWidget):
      QGridLayout.addWidget(self, widget, self.__total_count__, 0)
      self.__total_count__ += 1
      self.__row_count__ += 1

  def addGridWidget(self, *args) -> None:
    """Adds a widget to the layout. """
    widget = [*[a for a in args if isinstance(a, QWidget)], None][0]
    intArgs = [a for a in args if isinstance(a, int)]
    row, col, rowSpan, colSpan = None, None, None, None
    if len(intArgs) % 2:
      e = """Expected even number of integer arguments, but received: %d"""
      raise ValueError(e % len(intArgs))
    if len(intArgs) > 4:
      e = """Expected at most 4 integer arguments, but received: %d"""
      raise ValueError(e % len(intArgs))
    if len(intArgs) == 4:
      row, col, rowSpan, colSpan, = intArgs
    elif len(intArgs) == 2:
      row, col, rowSpan, colSpan, = intArgs[0], intArgs[1], 1, 1
    elif not intArgs:
      rowSpan, colSpan = 1, 1
      C, R = self.__col_count__, self.__row_count__
      c = self.__total_count__ % self.__col_count__
      r = self.__total_count__ % self.__row_count__
      if C > R:
        row, col = r - 1, c
        self.__col_count__ += colSpan
      else:
        row, col = r, c - 1
        self.__row_count__ += rowSpan
    if all([isinstance(a, int) for a in [row, col, rowSpan, colSpan]]):
      QGridLayout.addWidget(self, widget, row, col, rowSpan, colSpan)
      self.__total_count__ += (rowSpan * colSpan)

  def parentAddWidget(self, widget, *args) -> None:
    """Adds a widget to the layout. """
    widget.initUi()
    widget.connectActions()
    if isinstance(widget, BaseWidget):
      QGridLayout.addWidget(self, widget, *args)
    elif isinstance(widget, QWidget):
      QGridLayout.addWidget(self, widget, *args)
    else:
      e = typeMsg('widget', widget, BaseWidget)
      raise TypeError(e)


class Layout(BaseWidget):
  """Layout class with built-in widget"""

  baseWidget = AttriBox[BaseWidget]()
  innerLayout = AttriBox[InnerLayout]()
  __total_count__ = 0

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the Layout instance. """
    BaseWidget.__init__(self, )
    ic(self)
    self.__widget_registry__ = {}
    self.layoutType = None
    modeKeys = stringList("""layoutType, mode, orientation""")
    for key in modeKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if val in LayoutEnum:
          self.layoutType = val
          break
        if isinstance(val, str):
          if val in LayoutEnum:
            self.layoutType = LayoutEnum[val]
            break
    else:
      for arg in args:
        if arg in LayoutEnum:
          self.layoutType = arg
          break
      else:
        self.layoutType = LayoutEnum['grid']

  def addWidget(self, widget: BaseWidget, *args) -> None:
    """Adds a widget to the layout. """
    if self.innerLayout.__layout_mode__ == LayoutEnum.VERTICAL:
      return self.innerLayout.addVerticalWidget(widget)
    if self.innerLayout.__layout_mode__ == LayoutEnum.HORIZONTAL:
      return self.innerLayout.addHorizontalWidget(widget)
    row, col, rowSpan, colSpan = [*args, 1, 1, ][:4]
    key = row + col * 1j
    if key in self.__widget_registry__:
      e = """A widget already exists at the given position."""
      raise ValueError(e)
    self.innerLayout.parentAddWidget(widget, *args)
    self.__widget_registry__[key] = widget
    for i in range(rowSpan):
      for j in range(colSpan):
        self.__widget_registry__[key + i + j * 1j] = widget
        text = '__widget_%s_%s__' % (i, j)
        setattr(self, text, widget)

  def buildLayout(self) -> None:
    """Builds the layout. """
    for (position, widget) in self.__widget_registry__.items():
      if not isinstance(widget, (BaseWidget, QWidget)):
        e = typeMsg('widget', widget, BaseWidget)
        raise TypeError(e)
      widget.initUi()
      widget.connectActions()
      self.innerLayout.parentAddWidget(widget, *position)
    self.setLayout(self.innerLayout)
    self.initUi()



"""The 'settings' module centralizes default settings."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._default import Default
from ._icons import Icons



"""GridLayout widget implementation"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QGridLayout

from ezside.widgets import BoxLayout


class GridLayout(BoxLayout):
  """GridLayout widget."""

  def createBaseLayout(self) -> None:
    """Creates the base layout."""
    self.setBaseLayout(QGridLayout())

  def addWidget(self, widget: BoxLayout, row: int, column: int) -> None:
    """Adds a widget to the layout."""
    self.__child_widgets__.append(widget)
    widget.initUi()
    widget.connectActions()
    self.getBaseLayout().addWidget(widget, row, column)



"""StatusBar provides a status bar for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow, QStatusBar


class StatusBar(QStatusBar):
  """StatusBar provides a status bar for the main application window."""

  def initUi(self, ) -> None:
    """Initializes the user interface for the status bar."""
    self.showMessage('Ready')
    self.setStyleSheet("""background-color: #f0f0f0; color: #000000;
    border-top: 1px solid #000000; border-left: 1px solid #000000;""")



"""The 'widgets' package provides the widgets for the main application
window. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._timer import Timer
from ._data_roll import DataRoll
from ._data_series import DataSeries, LineSeries, ScatterSeries
from ._base_widget import BaseWidget
from ._spacer import VSpacer, HSpacer, Spacer, AbstractSpacer
from ._control_widget import HorizontalAbstractControl, AbstractControl
from ._control_widget import VerticalAbstractControl, PushButton
from ._text_label import TextLabel
from ._spinbox import SpinBox
from ._data_view import DataView
from ._control_widget import AbstractControl, PushButton
from ._plot_envelope import PlotEnvelope
from ._filled_rect import FilledRect
from ._vertical_panel import VerticalPanel
from ._horizontal_panel import HorizontalPanel
from ._corner_panel import CornerPanel
from ._client_info import ClientInfo
from ._vertical_slider import VerticalSlider
from ._equalizer import Equalizer
from ._white_noise import WhiteNoise
from ._data_widget import DataWidget



"""Equalizer provides a collection of vertical sliders."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time
from math import sin
from typing import Any, Self

from PySide6.QtCore import QSize
from PySide6.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout
from attribox import AttriBox
from vistutils.text import stringList, monoSpace
from vistutils.waitaminute import typeMsg

from ezside.core import Tight, Center
from ezside.widgets import VerticalSlider, BaseWidget, TextLabel, FilledRect, \
  HSpacer
from settings import Default


class Equalizer(BaseWidget):
  """Equalizer provides a collection of vertical sliders."""

  __vertical_sliders__ = None
  __iter_contents__ = None

  baseWidget = AttriBox[BaseWidget]()
  baseLayout = AttriBox[QHBoxLayout]()
  horizontalSpacer = AttriBox[HSpacer]()

  def __init__(self, *args, **kwargs) -> None:
    numSlidersKeys = stringList("""numSliders, num, sliders, numSliders""")
    templateKeys = stringList("""template, sliderTemplate, template""")
    Keys = [numSlidersKeys, templateKeys]
    names = stringList("""numSliders, templateSlider""")
    types = [int, QWidget]
    defaultValues = {
      names[0]: Default.numSliders,
      names[1]: VerticalSlider.defaultInstance()
    }
    data = {}
    for (name, type_, keys) in zip(names, types, Keys):
      for key in keys:
        if key in kwargs:
          val = kwargs.get(key)
          if isinstance(val, type_):
            data[name] = val
            break
      else:
        for arg in args:
          if isinstance(arg, type_):
            data[name] = arg
            break
        else:
          data[name] = defaultValues[name]
    self.numSliders = data.get(names[0])
    self._refSlider = data.get(names[1])
    BaseWidget.__init__(self, *args, **kwargs)

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(QSize(256, 256))
    self.setSizePolicy(Tight, Tight)
    for i in range(self.numSliders):
      vLayout = QVBoxLayout()
      vLayout.setContentsMargins(0, 0, 0, 0, )
      vWidget = FilledRect()
      sliderName = '__slider_%02d__' % i
      widgetName = '__widget_%02d__' % i
      layoutName = '__layout_%02d__' % i
      labelName = '__label_%02d__' % i
      slider = VerticalSlider.clone(self._refSlider)
      slider @= lambda val: val * sin(time.time() * 2 ** i)
      slider.initUi()
      label = TextLabel('%d' % 2 ** i)
      label.setSizePolicy(Tight, Tight)
      label.setMinimumSize(QSize(32, 32))
      vWidget = BaseWidget()
      setattr(self, sliderName, slider)
      setattr(self, labelName, label)
      setattr(self, widgetName, vWidget)
      setattr(self, layoutName, vLayout)
      vLayout.addWidget(label, alignment=Center)
      vLayout.addWidget(slider, alignment=Center)
      vWidget.setLayout(vLayout)
      self.baseLayout.addWidget(vWidget)

    self.baseLayout.addWidget(self.horizontalSpacer)
    self.setLayout(self.baseLayout)

  def __setitem__(self, index: int, title: str) -> None:
    """Set the title of the slider."""
    sliderName = '__slider_%02d__' % index
    labelName = '__label_%02d__' % index
    try:
      slider = getattr(self, sliderName)
      label = getattr(self, labelName)
      if not isinstance(slider, VerticalSlider):
        e = typeMsg(sliderName, slider, VerticalSlider)
        raise TypeError(e)
      if not isinstance(label, TextLabel):
        e = typeMsg(labelName, label, TextLabel)
        raise TypeError(e)
      label.setText(title)
      label.update()
      slider.update()
    except AttributeError as attributeError:
      e = """No slider at given index: '%d'""" % index
      raise IndexError(e) from attributeError
    except TypeError as typeError:
      e = """Unexpected type encountered at index: '%d'""" % index
      raise TypeError(e) from typeError

  def __getitem__(self, index: int) -> VerticalSlider:
    """Get the slider at the given index."""
    sliderName = '__slider_%02d__' % index
    try:
      return getattr(self, sliderName)
    except AttributeError as attributeError:
      e = """No slider at given index: '%d'""" % index
      raise IndexError(e) from attributeError

  def __matmul__(self, other: Any) -> Any:
    """Connect the signals of the two equalizers."""

    if isinstance(other, Equalizer):
      return self @ [float(slider) for slider in other]

    try:
      values = [val for val in other]
    except TypeError as typeError:
      if 'iterate' in str(typeError):
        return NotImplemented
      raise typeError
    if len(values) != self.numSliders:
      e = """Dimensional mismatch! Expected array of '%d' elements, but 
       received array of '%d' elements.""" % (self.numSliders, len(values))
      raise ValueError(monoSpace(e))
    if not isinstance(values, list):
      e = typeMsg('values', values, list)
      raise TypeError(e)
    out = 0j
    for (slider, value) in zip(self, values):
      out += complex(value) * float(slider)
    return out

  def __iter__(self) -> Self:
    """Iterate the sliders."""
    self.__iter_contents__ = [getattr(self, '__slider_%02d__' % i)
                              for i in range(self.numSliders)]
    return self

  def __next__(self, ) -> VerticalSlider:
    """Get the next slider."""
    if self.__iter_contents__ is None:
      e = """The iterator has not been initialized!"""
      raise RuntimeError(monoSpace(e))
    try:
      return self.__iter_contents__.pop(0)
    except IndexError:
      raise StopIteration

  def __abs__(self) -> float:
    """Return the absolute value of the equalizer."""
    return self @ [1. for _ in range(self.numSliders)]



"""DataSeries wraps the QXYSeries, QScatterSeries and QLineSeries classes
to provide a common interface for these classes."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import numpy as np
from PySide6.QtCharts import QXYSeries, QLineSeries, QScatterSeries
from PySide6.QtCore import Slot
from attribox import AttriBox

from ezside.widgets import DataRoll


class DataSeries(QXYSeries):
  """DataSeries wraps the QXYSeries, QScatterSeries and QLineSeries classes
  to provide a common interface for these classes."""

  data = AttriBox[DataRoll]()

  def updateValues(self, ) -> None:
    """Updates the values of the series."""
    data = self.data.snapShot()
    t = data.real
    x = data.imag
    self.clear()
    self.appendNp(t.astype(np.float32), x.astype(np.float32))


class ScatterSeries(QScatterSeries):
  """ScatterSeries wraps the QScatterSeries class to provide a common
  interface for the QXYSeries, QScatterSeries and QLineSeries classes."""

  data = AttriBox[DataRoll]()

  def updateValues(self, ) -> None:
    """Updates the values of the series."""
    data = self.data.snapShot()
    t = data.real
    x = data.imag
    self.clear()
    self.appendNp(t.astype(np.float32), x.astype(np.float32))

  @Slot(float)
  def appendValue(self, value: float) -> None:
    """Appends a value to the series."""
    self.data.append(value)


class LineSeries(QLineSeries):
  """ScatterSeries wraps the QScatterSeries class to provide a common
  interface for the QXYSeries, QScatterSeries and QLineSeries classes."""

  data = AttriBox[DataRoll]()

  def updateValues(self, ) -> None:
    """Updates the values of the series."""
    data = self.data.snapShot()
    t = data.real
    x = data.imag
    self.clear()
    self.appendNp(t.astype(np.float32), x.astype(np.float32))

  @Slot(float)
  def appendValue(self, value: float) -> None:
    """Appends a value to the series."""
    self.data.append(value)



"""IntField is a strongly typed descriptor class"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from vistutils.text import monoSpace


class IntField:
  """IntField provides a strongly typed descriptor field for integers"""

  __field_name__ = None
  __field_owner__ = None
  __fallback_value__ = 0

  def __set_name__(self, owner, name) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner

  def _getPrivateFieldName(self) -> str:
    """Getter-function for the private field name"""
    return '_%s' % self.__field_name__

  def __get__(self, instance, owner, **kwargs) -> None:
    pvtName = self._getPrivateFieldName()
    if getattr(instance, pvtName, None) is None:
      setattr(instance, pvtName, self.__fallback_value__)
      return self.__get__(instance, owner, _recursion=True)
    return getattr(instance, pvtName)

  def __set__(self, instance, value) -> None:
    """Setter-function for the descriptor"""
    if isinstance(value, int):
      pvtName = self._getPrivateFieldName()
      return setattr(instance, pvtName, value)

  def __delete__(self, instance) -> None:
    """Deleter-function for the descriptor"""
    pvtName = self._getPrivateFieldName()
    if hasattr(instance, pvtName):
      return delattr(instance, pvtName)
    e = """No attribute found at name: '%s'""" % (pvtName,)
    raise AttributeError(monoSpace(e))



"""This package provides a central registry of the keyboard shortcuts used
by the main application."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._get_shortcut import getShortcut



"""WrapView wraps the QChartView subclasses, as they must remain
subclasses of QGraphicView. The WrapView widget shows the DataView
instance while remaining a subclass of BaseWidget. This allows the
DataView to be used in layouts and other widgets that require a BaseWidget
instance."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Slot
from attribox import AttriBox, this
from icecream import ic
from vistutils.parse import maybe

from ezside.widgets import BaseWidget, DataView,

ic.configureOutput(includeContext=True)


class WrapView(BaseWidget):
  """WrapView provides a wrapper around the DataView."""

  __this_vals__ = None

  dataView = AttriBox[DataView](this)
  baseLayout = AttriBox[Layout]('horizontal')

  def __init__(self, seriesType: str = None, *args, **kwargs):
    BaseWidget.__init__(self, *args, **kwargs)
    self._seriesType = maybe(seriesType, 'scatter')

  def initUi(self) -> None:
    """Initializes the user interface for the main window."""
    self.baseLayout.addWidget(self.dataView)
    self.setLayout(self.baseLayout)

  @Slot()
  def refresh(self) -> None:
    """Refreshes the data."""

  def getSeriesType(self) -> str:
    """Returns the series type."""
    return self._seriesType or 'scatter'



"""Noisinator creates a noise signal."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Signal
from PySide6.QtGui import QPaintEvent, QPainter
from PySide6.QtWidgets import QHBoxLayout, QVBoxLayout
from attribox import AttriBox
from icecream import ic

from moreutils import IntField
from ezside.core import Precise, LawnGreen
from ezside.widgets import TextLabel, Equalizer, BaseWidget, \
  VerticalAbstractControl
from ezside.widgets import Timer
from settings import Default


class WhiteNoise(BaseWidget):
  """Noisinator creates a noise signal."""

  noise = Signal(float)
  start = Signal()
  stop = Signal()
  pause = Signal()

  resumeTimer = IntField()

  timer = AttriBox[Timer](Default.noiseTimer, Precise, False)
  baseLayout = AttriBox[QVBoxLayout]()
  horizontalWidget = AttriBox[BaseWidget]()
  horizontalLayout = AttriBox[QHBoxLayout]()
  equalizer = AttriBox[Equalizer](6, )
  control = AttriBox[VerticalAbstractControl]()
  titleBanner = AttriBox[TextLabel]("""Let's make some noise!""")

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(400, 400)
    self.equalizer.initUi()
    self.horizontalLayout.addWidget(self.equalizer)
    self.control.initUi()
    self.horizontalLayout.addWidget(self.control)
    self.horizontalWidget.setLayout(self.horizontalLayout)
    self.baseLayout.addWidget(self.horizontalWidget)
    self.setLayout(self.baseLayout)
    self.connectActions()

  def connectActions(self) -> None:
    """The connectActions method connects the actions."""
    self.control.start.connect(self.start)
    self.control.stop.connect(self.stop)
    self.control.pause.connect(self.pause)
    self.start.connect(self.startHandle)
    self.stop.connect(self.stopHandle)
    self.pause.connect(self.pauseHandle)
    self.timer.timeout.connect(self.emitNoise)

  def startHandle(self) -> None:
    """Handles the start signal"""
    ic()
    if self.resumeTimer:
      timer = Timer(self.resumeTimer, Precise, True)
      timer.timeout.connect(self.timer.start)
      return timer.start()
    self.timer.start()

  def stopHandle(self) -> None:
    """Handles the stop signal"""
    self.resumeTimer = 0
    self.timer.stop()

  def pauseHandle(self) -> None:
    """Handles the pause signal"""
    self.resumeTimer = self.timer.remainingTime()
    self.timer.stop()

  def emitNoise(self, ) -> None:
    """Emit the noise signal."""
    self.noise.emit(float(self))

  def __float__(self) -> float:
    """The float method returns the float representation of the object."""
    if isinstance(self.equalizer, Equalizer):
      return abs(self.equalizer).real
    return NotImplemented

  def paintEvent(self, event: QPaintEvent) -> None:
    """The paintEvent method paints the widget."""
    painter = QPainter()
    painter.begin(self)
    rect = painter.viewport()
    painter.setRenderHint(QPainter.Antialiasing)
    self.solidBrush.setColor(LawnGreen)
    painter.setBrush(self.solidBrush)
    painter.setPen(self.emptyLine)
    painter.drawRoundedRect(rect, 2, 2)
    painter.end()



"""DebugMenu provides actions for debugging the main application."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow
from icecream import ic

from attribox import AttriBox
from ezside.windows.menus import AbstractMenu, Action

ic.configureOutput(includeContext=True, )


class DebugMenu(AbstractMenu):
  """DebugMenu provides actions for debugging the main application."""

  debug1 = AttriBox[Action]('debug1')
  debug2 = AttriBox[Action]('debug2')
  debug3 = AttriBox[Action]('debug3')
  debug4 = AttriBox[Action]('debug4')
  debug5 = AttriBox[Action]('debug5')
  debug6 = AttriBox[Action]('debug6')
  debug7 = AttriBox[Action]('debug7')
  debug8 = AttriBox[Action]('debug8')
  debug9 = AttriBox[Action]('debug9')

  def __init__(self, parent: QMainWindow) -> None:
    """DebugMenu provides actions for debugging the main application."""
    AbstractMenu.__init__(self, parent)
    self.setTitle('Debug')

  def appendActions(self) -> None:
    """Appends the actions without setting them up"""
    self.addAction(self.debug1)
    self.addAction(self.debug2)
    self.addAction(self.debug3)
    self.addSeparator()
    self.addAction(self.debug4)
    self.addAction(self.debug5)
    self.addAction(self.debug6)
    self.addSeparator()
    self.addAction(self.debug7)
    self.addAction(self.debug8)
    self.addAction(self.debug9)

  def setupActions(self, ) -> None:
    """Sets up the actions for the debug menu."""
    self.debug1.setupAction()
    self.debug2.setupAction()
    self.debug3.setupAction()
    self.debug4.setupAction()
    self.debug5.setupAction()
    self.debug6.setupAction()
    self.debug7.setupAction()
    self.debug8.setupAction()
    self.debug9.setupAction()



"""HorizontalPanel provides the same as VerticalPanel, but in a horizontal
direction."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QColor
from vistutils.parse import maybe

from ezside.core import Fixed, Expand, LawnGreen
from ezside.widgets import FilledRect


class HorizontalPanel(FilledRect):
  """HorizontalPanel provides the same as VerticalPanel, but in a horizontal
  direction."""

  def __init__(self, *args, **kwargs) -> None:
    colorArgs, intArgs, width = [], [], None
    color = None
    for arg in args:
      if isinstance(arg, int):
        intArgs.append(arg)
      elif isinstance(arg, QColor):
        colorArgs.append(arg)
    if colorArgs:
      color = None
    if color is None:
      if len(intArgs) == 3:
        color = QColor(*intArgs, 255)
      elif len(intArgs) > 3:
        color = QColor(*intArgs[:4])
    color = maybe(color, LawnGreen)
    self.fillColor.setRed(color.red())
    self.fillColor.setGreen(color.green())
    self.fillColor.setBlue(color.blue())
    FilledRect.__init__(self, *args, **kwargs)
    self.setFixedHeight(32)
    self.setMaximumWidth(1024)
    self.setSizePolicy(Fixed, Expand)



"""FilesMenu provides the file menu for the main application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow
from icecream import ic

from attribox import AttriBox
from ezside.windows.menus import AbstractMenu, Action

ic.configureOutput(includeContext=True, )


class FilesMenu(AbstractMenu):
  """FilesMenu provides the file menu for the main application window."""

  new = AttriBox[Action]('new')
  open = AttriBox[Action]('open')
  save = AttriBox[Action]('save')
  save_as = AttriBox[Action]('save_as')
  exit = AttriBox[Action]('exit')

  def __init__(self, parent: QMainWindow) -> None:
    """FilesMenu provides the file menu for the main application window."""
    AbstractMenu.__init__(self, parent)
    self.setTitle('File')

  def appendActions(self) -> None:
    """Appends the actions without setting them up"""
    self.addAction(self.new)
    self.addAction(self.open)
    self.addAction(self.save)
    self.addAction(self.save_as)
    Action(self.addSeparator())
    self.addAction(self.exit)

  def setupActions(self, ) -> None:
    """Sets up the actions for the file menu."""
    self.new.setupAction()
    self.open.setupAction()
    self.save.setupAction()
    self.save_as.setupAction()
    self.exit.setupAction()



"""The auto singleton should be used to include the name as an enum
instance"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from attribox import Singleton


class auto(metaclass=Singleton):
  """LMAO"""

  def __call__(self, *args, **kwargs) -> type:
    """This is crazy!"""
    return auto



"""The getIcon function creates and returns a QIcon of the given name."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from icecream import ic

ic.configureOutput(includeContext=True)


def getIconPath(name: str) -> str:
  """Creates and returns a QIcon of the given name."""

  name = name.lower().strip()
  name = name.replace(' ', '_')
  here = os.path.dirname(__file__)
  baseName = name.split('.')[0]
  iconPath = os.path.join(here, baseName + '.png')
  return iconPath



"""BaseWindow provides the base class for the main application window. It
implements menus and actions for the application, leaving widgets for the
LayoutWindow class."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QMainWindow, QApplication

from attribox import AttriBox, this
from ezside.windows.bars import MenuBar, StatusBar


class BaseWindow(QMainWindow):
  """BaseWindow provides the base class for the main application window. It
  implements menus and actions for the application, leaving widgets for the
  LayoutWindow class."""

  mainMenuBar = AttriBox[MenuBar](this)
  mainStatusBar = AttriBox[StatusBar](this)

  def initUi(self, ) -> None:
    """Initializes the user interface for the main window."""
    self.mainMenuBar.initUi()
    self.setMenuBar(self.mainMenuBar)
    self.mainStatusBar.initUi()
    self.setStatusBar(self.mainStatusBar)
    self.connectActions()

  def connectActions(self) -> None:
    """Connects the actions to the business logic."""
    self.mainMenuBar.debug.debug1.triggered.connect(self.debug1Func)
    self.mainMenuBar.debug.debug2.triggered.connect(self.debug2Func)
    self.mainMenuBar.debug.debug3.triggered.connect(self.debug3Func)
    self.mainMenuBar.addSeparator()
    self.mainMenuBar.debug.debug4.triggered.connect(self.debug4Func)
    self.mainMenuBar.debug.debug5.triggered.connect(self.debug5Func)
    self.mainMenuBar.debug.debug6.triggered.connect(self.debug6Func)
    self.mainMenuBar.addSeparator()
    self.mainMenuBar.debug.debug7.triggered.connect(self.debug7Func)
    self.mainMenuBar.debug.debug8.triggered.connect(self.debug8Func)
    self.mainMenuBar.debug.debug9.triggered.connect(self.debug9Func)
    self.mainMenuBar.help.about_qt.triggered.connect(QApplication.aboutQt)
    self.mainMenuBar.files.exit.triggered.connect(self.close)

  def debug1Func(self, ) -> None:
    """Debug1 function."""
    note = 'Debug1 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug2Func(self, ) -> None:
    """Debug2 function."""
    note = 'Debug2 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug3Func(self, ) -> None:
    """Debug3 function."""
    note = 'Debug3 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug4Func(self, ) -> None:
    """Debug4 function."""
    note = 'Debug4 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug5Func(self, ) -> None:
    """Debug5 function."""
    note = 'Debug5 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug6Func(self, ) -> None:
    """Debug6 function."""
    note = 'Debug6 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug7Func(self, ) -> None:
    """Debug7 function."""
    note = 'Debug7 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug8Func(self, ) -> None:
    """Debug8 function."""
    note = 'Debug8 function called'
    print(note)
    self.statusBar().showMessage(note)

  def debug9Func(self, ) -> None:
    """Debug9 function."""
    note = 'Debug9 function called'
    print(note)
    self.statusBar().showMessage(note)



"""Spacer widget takes up space"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import QMargins
from PySide6.QtGui import QPaintEvent, QPainter, QPen, QColor, QBrush

from ezside.core import Expand, Tight, White, DashLine
from ezside.widgets import BaseWidget


class AbstractSpacer(BaseWidget):
  """Spacer widget takes up space"""

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the widget."""
    self.initUi()
    self.connectActions()

  def connectActions(self) -> None:
    """The connectActions method connects the widget actions."""
    pass

  def paintEvent(self, event: QPaintEvent) -> None:
    """The paintEvent method is called when the widget needs to be
    repainted."""
    painter = QPainter()
    painter.begin(self)
    viewRect = painter.viewport()
    paintRect = viewRect - QMargins(6, 6, 6, 6)
    paintRect.moveCenter(viewRect.center())
    debugBrush = QBrush()
    debugBrush.setColor(QColor(255, 255, 255, 31))
    debugPen = QPen()
    debugPen.setColor(White)
    debugPen.setStyle(DashLine)
    painter.setBrush(debugBrush)
    painter.setPen(debugPen)
    painter.drawRoundedRect(painter.viewport(), 4, 4)
    painter.end()


class VSpacer(AbstractSpacer):
  """Spacer widget takes up space"""

  def __init__(self, *args, **kwargs) -> None:
    AbstractSpacer.__init__(self, *args, **kwargs)
    self.setSizePolicy(Tight, Expand)


class HSpacer(AbstractSpacer):
  """Spacer widget takes up space"""

  def __init__(self, *args, **kwargs) -> None:
    AbstractSpacer.__init__(self, *args, **kwargs)
    self.setSizePolicy(Expand, Tight)


class Spacer(AbstractSpacer):
  """Spacer widget takes up space"""

  def __init__(self, *args, **kwargs) -> None:
    AbstractSpacer.__init__(self, *args, **kwargs)
    self.setSizePolicy(Expand, Expand)



"""StrField provides a strongly typed descriptor field for strings"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from vistutils.parse import maybe
from vistutils.text import monoSpace
from vistutils.waitaminute import typeMsg


class StrField:
  """StrField provides a strongly typed descriptor field for strings"""
  __field_name__ = None
  __field_owner__ = None
  __fallback_value__ = ''
  __default_value__ = None

  def __init__(self, defVal: str = None) -> None:
    text = maybe(defVal, self.__fallback_value__)
    if isinstance(text, str):
      self.__default_value__ = text
    else:
      e = typeMsg('defVal', defVal, str)
      raise TypeError(e)

  def __set_name__(self, owner: type, name: str) -> None:
    self.__field_name__ = name
    self.__field_owner__ = owner

  def getDefaultValue(self) -> str:
    """Getter-function for the default value"""
    return self.__default_value__

  def getFieldOwner(self) -> type:
    """Getter-function for the field owner"""
    if self.__field_owner__ is not None:
      if isinstance(self.__field_owner__, type):
        return self.__field_owner__
      e = typeMsg('self.__field_owner__', self.__field_owner__, type)
      raise TypeError(e)
    e = """Instances of StrField, and descriptors in general, will not 
    have their field name populated until their owning class is actually 
    created. During the execution of the owning class body, objects are 
    created and are placed in the namespace object created by the 
    metaclass (normally type which uses a plain 'dict' object). This error 
    indicates an attempt has been made to access the field name before the 
    metaclass has finished creating the owning class. """
    raise AttributeError(monoSpace(e))

  def getFieldName(self) -> str:
    """Getter-function for the private field name"""
    if self.__field_name__ is not None:
      if isinstance(self.__field_name__, str):
        return self.__field_name__
      e = typeMsg('self.__field_name__', self.__field_name__, str)
      raise TypeError(e)
    e = """Instances of StrField, and descriptors in general, will not 
    have their field name populated until their owning class is actually 
    created. During the execution of the owning class body, objects are 
    created and are placed in the namespace object created by the 
    metaclass (normally type which uses a plain 'dict' object). This error 
    indicates an attempt has been made to access the field name before the 
    metaclass has finished creating the owning class. """
    raise AttributeError(monoSpace(e))

  def _getPrivateFieldName(self, ) -> str:
    """Getter-function for the private field name"""
    return '_%s' % self.getFieldName()

  def __get__(self, instance: object, owner: type, **kwargs) -> Any:
    """Implementation of the getter function for the descriptor"""
    if instance is None:
      return self
    pvtName = self._getPrivateFieldName()
    if getattr(instance, pvtName, None) is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      setattr(instance, pvtName, self.getDefaultValue())
      return self.__get__(instance, owner, _recursion=True)
    return getattr(instance, pvtName)

  def __set__(self, instance: object, value: str) -> None:
    """Implementation of the setter function for the descriptor"""
    if isinstance(value, str):
      pvtName = self._getPrivateFieldName()
      return setattr(instance, pvtName, value)
    e = typeMsg('value', value, str)
    raise TypeError(e)

  def __delete__(self, instance: object) -> None:
    """Implementation of the deleter function for the descriptor"""
    pvtName = self._getPrivateFieldName()
    if hasattr(instance, pvtName):
      return delattr(instance, pvtName)
    e = """No attribute found at name: '%s'""" % (pvtName,)
    raise AttributeError(monoSpace(e))

  def __str__(self, ) -> str:
    """Implementation of the string function for the descriptor"""
    try:
      fieldName = self.getFieldName()
      ownerName = self.getFieldOwner()
      return '%s.%s' % (ownerName, fieldName)
    except AttributeError as attributeError:
      if 'During the execution of the owning class ' in str(attributeError):
        return object.__str__(self)
      raise attributeError

  def __repr__(self, ) -> str:
    """Implementation of the representation function for the descriptor"""
    clsName = self.__class__.__name__
    if self.__default_value__ == self.__fallback_value__:
      return '%s()' % clsName
    defVal = self.getDefaultValue()
    return '%s(%s)' % (clsName, defVal)



"""BaseWidget provides a base class for the widgets. Using AttriBox they
provide brushes, pens and fonts as attributes. These widgets are not meant
for composite widgets directly but instead for the constituents. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt, QSize
from PySide6.QtGui import QColor, QPainter
from PySide6.QtWidgets import QWidget
from attribox import AttriBox
from icecream import ic

from ezside.core import parseParent, NoWrap, Center, Pen, Font, Brush
from ezside.core import BevelJoin, FlatCap, SolidLine, SolidFill, DashLine
from ezside.core import DotLine, DashDot

ic.configureOutput(includeContext=True, )


class BaseWidget(QWidget):
  """BaseWidget provides a base class for the widgets. Using AttriBox they
  provide brushes, pens and fonts as attributes. These widgets are not meant
  for composite widgets directly but instead for the constituents. """

  baseSize = AttriBox[QSize](32, 32)
  solidBrush = AttriBox[Brush](SolidFill)
  emptyBrush = AttriBox[Brush](QColor(0, 0, 0, 0))
  solidLine = AttriBox[Pen](SolidLine)
  dashedLine = AttriBox[Pen](DashLine)
  dottedLine = AttriBox[Pen](DotLine)
  dashDotLine = AttriBox[Pen](DashDot)
  fontLine = AttriBox[Pen](SolidLine, 1, FlatCap, BevelJoin)
  emptyLine = AttriBox[Pen](Qt.PenStyle.NoPen)
  defaultFont = AttriBox[Font]('Montserrat', 16, )

  def __init__(self, *args, **kwargs) -> None:
    """BaseWidget provides a base class for the widgets. Using AttriBox they
    provide brushes, pens and fonts as attributes. These widgets are not
    meant for composite widgets directly but instead for the components."""
    # ic('%s' % self.__class__.__name__)
    parent = parseParent(*args, **kwargs)
    QWidget.__init__(self, parent)
    self.setMinimumSize(QSize(64, 64))
    # self.initUi()

  def painterPrint(self, painter: QPainter) -> QPainter:
    """The painterPrint method adjusts the given painter to print text."""
    painter.setPen(self.fontLine)
    painter.setFont(self.defaultFont)
    return painter

  def painterFill(self, painter: QPainter) -> QPainter:
    """The painterFill method adjusts the given painter to fill shapes."""
    painter.setPen(self.emptyLine)
    painter.setBrush(self.solidBrush)
    return painter

  def painterLine(self, painter: QPainter) -> QPainter:
    """The painterLine method adjusts the given painter to draw lines."""
    painter.setPen(self.solidLine)
    painter.setBrush(self.emptyBrush)
    return painter

  def boundSize(self, text: str, ) -> None:
    """The boundSize method returns the bounding rectangle of the given
    text."""
    rect, flags = self.geometry(), NoWrap | Center
    self.defaultFont.metrics().boundingRect(flags, flags, text)
    return rect.size()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(self.baseSize)
    self.connectActions()

  def connectActions(self) -> None:
    """The connectActions method connects the actions to the signals."""



"""AbstractMenu provides the base class for the custom menus. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod

from PySide6.QtCore import QEvent
from PySide6.QtWidgets import QMenu, QMainWindow
from icecream import ic

from ezside.windows.menus import Action

ic.configureOutput(includeContext=True, )


class AbstractMenu(QMenu):
  """AbstractMenu provides the base class for the custom menus. """

  def __init__(self, parent: QMainWindow) -> None:
    """Initializes the menu."""
    QMenu.__init__(self, parent)

  @abstractmethod
  def appendActions(self) -> None:
    """Appends the actions without setting them up"""

  @abstractmethod
  def setupActions(self, ) -> None:
    """Sets up the actions for the menu."""

  def applyParent(self) -> None:
    """Sets the parent on actions"""
    for action in self.actions():
      if isinstance(action, Action):
        action.setParent(self.parent())

  def initUi(self) -> None:
    """Collections append, setup and apply parent"""
    self.appendActions()
    self.setupActions()
    self.applyParent()



"""Pen wraps the QPen class."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt
from PySide6.QtGui import QPen, QColor
from vistutils.parse import maybe

from ezside.core import BevelJoin, FlatCap, SolidLine, Black


class Pen(QPen):
  """Pen wraps the QPen class."""

  def __init__(self, *args, **kwargs) -> None:
    QPen.__init__(self, )
    intArgs = [arg for arg in args if isinstance(arg, int)]
    colorArgs = [arg for arg in args if isinstance(arg, QColor)]
    styleArgs = [arg for arg in args if isinstance(arg, Qt.PenStyle)]
    capStyleArgs = [arg for arg in args if isinstance(arg, Qt.PenCapStyle)]
    joinStyleArgs = [arg for arg in args if isinstance(arg, Qt.PenJoinStyle)]
    width = None
    color = None
    if not colorArgs:
      if len(intArgs) in [4, 5]:
        width = intArgs.pop(0)
        color = QColor(*intArgs)
      elif len(intArgs) > 5:
        raise ValueError
      elif intArgs:
        width = 1
    color = maybe(color, Black)
    width = maybe(width, 1)
    penStyle = styleArgs[0] if styleArgs else SolidLine
    capStyle = capStyleArgs[0] if capStyleArgs else FlatCap
    joinStyle = joinStyleArgs[0] if joinStyleArgs else BevelJoin
    self.setWidth(width)
    self.setColor(color)
    self.setStyle(penStyle)
    self.setCapStyle(capStyle)
    self.setJoinStyle(joinStyle)



"""LayoutWindow subclasses BaseWindow and implements the layout of
widgets."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QGridLayout
from attribox import AttriBox
from icecream import ic

from ezside.core import LawnGreen
from ezside.widgets import (HorizontalPanel, DataWidget, ClientInfo,
                            BaseWidget, \
                            WhiteNoise, PlotEnvelope)
from ezside.widgets import TextLabel, CornerPanel, VerticalPanel
from ezside.windows import BaseWindow

ic.configureOutput(includeContext=True, )


class LayoutWindow(BaseWindow):
  """LayoutWindow subclasses BaseWindow and implements the layout of
  widgets."""

  baseWidget = AttriBox[BaseWidget]()
  baseLayout = AttriBox[QGridLayout]()

  left = AttriBox[VerticalPanel](LawnGreen)
  top = AttriBox[HorizontalPanel](LawnGreen)
  right = AttriBox[VerticalPanel](LawnGreen)
  bottom = AttriBox[HorizontalPanel](LawnGreen)
  topLeft = AttriBox[CornerPanel](LawnGreen)
  topRight = AttriBox[CornerPanel](LawnGreen)
  bottomLeft = AttriBox[CornerPanel](LawnGreen)
  bottomRight = AttriBox[CornerPanel](LawnGreen)

  titleBanner = AttriBox[TextLabel]('EZside')
  whiteNoise = AttriBox[WhiteNoise]()
  clientInfo = AttriBox[ClientInfo]()
  dataWidget = AttriBox[DataWidget]()
  plotEnvelope = AttriBox[PlotEnvelope]()

  def initUi(self) -> None:
    """The initUi method initializes the user interface of the window."""
    self.setMinimumSize(400, 400)
    rowCount = 2
    colCount = 2

    self.bottomRight.initUi()
    self.bottomLeft.initUi()
    self.topRight.initUi()
    self.topLeft.initUi()
    self.left.initUi()
    self.top.initUi()
    self.right.initUi()
    self.bottom.initUi()

    self.baseLayout.addWidget(self.bottomRight, rowCount + 1, 1 + colCount)
    self.baseLayout.addWidget(self.bottom, rowCount + 1, 1, 1, colCount)
    self.baseLayout.addWidget(self.bottomLeft, rowCount + 1, 0)

    self.baseLayout.addWidget(self.left, 1, 0, rowCount, 1)
    self.baseLayout.addWidget(self.right, 1, colCount + 1, rowCount, 1)

    self.baseLayout.addWidget(self.topLeft, 0, 0)
    self.baseLayout.addWidget(self.top, 0, 1, 1, colCount)
    self.baseLayout.addWidget(self.topRight, 0, colCount + 1)

    self.whiteNoise.initUi()
    self.baseLayout.addWidget(self.whiteNoise, 1, 2)
    self.plotEnvelope.initUi()
    self.baseLayout.addWidget(self.plotEnvelope, 2, 2)
    self.dataWidget.initUi()
    self.baseLayout.addWidget(self.dataWidget, 1, 1, 2, 1)
    self.baseWidget.setLayout(self.baseLayout)
    self.setCentralWidget(self.baseWidget)
    BaseWindow.initUi(self, )



"""The icons module provides a collection of icons for use in the
application. The icons are provided as a dictionary with the icon name
as the key and the icon path as the value."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os
from typing import Any

from PIL import Image
from PIL import ImageQt
from PySide6.QtGui import QIcon


class Icons:
  """The icons class provides icons"""

  @staticmethod
  def _getIconPath() -> str:
    """Returns the path to the icon folder."""
    fileDir = os.path.join(os.path.dirname(__file__), '../')
    iconPath = os.path.join(fileDir, 'ezside', 'windows', 'menus', 'icons', )
    return os.path.normpath(iconPath)

  @classmethod
  def _getIconList(cls, ) -> Image:
    """Returns the icon with the given name."""
    iconPath = cls._getIconPath()
    return os.listdir(iconPath)

  @classmethod
  def load(cls, name: str) -> Any:
    """Loads the icons from the icon folder."""
    iconList = cls._getIconList()
    for icon in iconList:
      if name in icon:
        iconFile = os.path.join(cls._getIconPath(), icon)
        img = Image.open(iconFile)
        return QIcon(ImageQt.toqpixmap(img))
    else:
      return cls.load('risitas')



"""The 'ezside.windows.menus' package provides the menus for the main
application window."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._action import Action
from ._abstract_menu import AbstractMenu
from ._files_menu import FilesMenu
from ._edit_menu import EditMenu
from ._help_menu import HelpMenu
from ._debug_menu import DebugMenu
